/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
// sga.cc: Main program for SGA (Simple GIF Animator)

#define DEBUG_SUMMARY 0
#define DEBUG_EXC 0

#if DEBUG_EXC
void checkHeap(char *title);
void dumpHeap(char *title);
#endif

// Define EXTERN to be blank so global variables will be allocated here
#define EXTERN

#define GIF_DIR_MASK "*.gif"
#if 0
// This is temporary
#ifdef WIN32
#define GIF_DIR_MASK "*.gif"
#else
#define GIF_DIR_MASK "gif/""*.gif"
#endif
#endif

#define WELCOME "Welcome to SGA"

#define TITLE_MAX 80

#define N_TEXT_ROWS 14
#define N_TEXT_ROWS_SERVER 32
#define N_TEXT_COLS 40

#define LIST_WIDTH 100
#define LIST_VISIBLE 10

#define GRAPH_HEIGHT 400
#define GRAPH_WIDTH 400

// mainCB
#define QUIT_BTN                 0
#define FILE_OPEN_BTN            1
#define FILE_NEW_BTN             2
#define FILE_SAVEAS_BTN          3
#define FILE_SAVE_BTN            4
#define INSERT_GIF_BTN           5
#define INSERT_COMMENT_BTN       6
#define PLAY_BTN                 7
#define STOP_BTN                 8
#define FIRST_BTN                9
#define NEXT_BTN                 10
#define PREV_BTN                 11
#define LAST_BTN                 12
#define DELETE_BTN               13
#define MOVE_UP_BTN              14
#define MOVE_DOWN_BTN            15
#define TOOLS_OPTIONS_BTN        16
#define OPTIONS_CANCEL_BTN       17

// optionsCB
#define OPTIONS_DISPLAY_MODE_BTN   100
#define OPTIONS_BACKGROUND_BTN     101
#define OPTIONS_MERGE_MODE_BTN     102
#define OPTIONS_DEF_COLORTABLE_BTN 103

// optCB
#define APPLY_ALL_BTN               0
#define TRANSPARENT_COLOR_BTN       1
#define INTERLACE_BTN               2
#define LOCAL_COLOR_TABLE_BTN       3
#define TRANSPARENT_INDEX_ENTRY     4
#define DELAY_TIME_ENTRY            5
#define LEFT_OFFSET_ENTRY           6
#define TOP_OFFSET_ENTRY            7
#define BG_COLOR_INDEX_ENTRY        8
#define LOGICAL_SCREEN_WIDTH_ENTRY  9
#define LOGICAL_SCREEN_HEIGHT_ENTRY 10
#define COMMENT_APPLY_BTN           11

// colorTableCB
#define COLOR_TABLE_EDIT_BTN    1000
#define COLOR_TABLE_BTN         1000
#define COLOR_TABLE_APPLY_BTN   1001
#define COLOR_TABLE_CLOSE_BTN   1002
#define COLOR_TABLE_TRIM_BTN    1003
#define COLOR_TABLE_UNUSED_BTN  1004
#define COLOR_TABLE_INDEX_ENTRY 1005
#define COLOR_TABLE_RED_ENTRY   1006
#define COLOR_TABLE_GREEN_ENTRY 1007
#define COLOR_TABLE_BLUE_ENTRY  1008
#define COLOR_TABLE_RED_SCALE   1009
#define COLOR_TABLE_GREEN_SCALE 1010
#define COLOR_TABLE_BLUE_SCALE  1011

#define DO_RED   0
#define DO_GREEN 1
#define DO_BLUE  2

#define MATCH_MODE       0
#define REPLACE_MODE     1
#define ADD_MODE         2
#define ADDREPLACE_MODE  3

#define N_DEFAULT_COLORS 8
#define DEFAULT_COLOR_N  3

#include "CXpm.h"
#include "CSGA.h"

#include <time.h>
#include <sys/stat.h>
#include <errno.h>

#ifdef WIN32
/* WIN32 does not have unistd.h and does not define the following constants */
#define F_OK 00
#define W_OK 02
#define R_OK 04
#include <direct.h>     /* for getcwd (usually in sys/parm.h or unistd.h) */
#include <io.h>         /* for access, chmod  (usually in unistd.h) */
#else
/* Use unistd.h */
#include <unistd.h>
#endif

#include <X11/Xmu/Editres.h>
#include <X11/cursorfont.h>
#include <Xm/ArrowB.h>
#include <Xm/CascadeB.h>
#include <Xm/CascadeBG.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/MainW.h>
#include <Xm/MessageB.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/SelectioB.h>
#include <Xm/Separator.h>
#include <Xm/Scale.h>
#include <Xm/ScrolledW.h>
#include <Xm/TextF.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>

#include <Xm/MwmUtil.h>
#include <Xm/Protocols.h>
#include <X11/Xatom.h>

// Fallback resourses
#include "resources.h"

// Define help text
#include "sga.hlp"

// Logo
//#include "epics.xpm"
#include "sga.xpm"

/* Structures */

/* Function prototypes */

int main(int argc, char **argv);
int clearListArea(void);
int clearDisplayArea(const char *text);
void colorTableCB(Widget w, XtPointer clientData, XtPointer callData);
void controlCB(Widget w, XtPointer clientData, XtPointer callData);
void createColorTablePopup(void);
static int createOptionsDialog(void);
Widget createCommentOptionArea(Widget parent);
Widget createDefaultOptionArea(Widget parent);
Widget createGlobalOptionArea(Widget parent);
Widget createImageOptionArea(Widget parent);
void createMainFormAreas(Widget wparent);
void displayMsg(char *fmt, ...);
Pixmap getLogo(Pixel bg, char **xpm);
void helpCB(Widget, XtPointer clientData, XtPointer);
int insertFile(char *filename);
int loadFile(char *filename);
void mainCB(Widget w, XtPointer clientData, XtPointer callData);
static void optionsCB(Widget w, XtPointer clientData, XtPointer callData);
static void optionMenuSet(Widget menu, int buttonId);
int parseCommand(int argc, char **argv);
void playTimerCB(XtPointer cd, XtIntervalId *id);
void popupColorTable(void);
static int popupOptionsDialog(void);
void print(const char *fmt, ...);
void printMsg(void);
void quit(Widget w, XtPointer clientData, XtPointer callData);
void refreshColorTableS(void);
int saveFile(char *fileName);
void setColorTableIndex(int index);
int setWindowTitle(char * fileName);
void startPlay(void);
void stopPlay(void);
char *timeStamp(void);
void updateColorTableSample(int component,int value);
void usage(void);
static void executeFSD(void);
static void fileSelectCB(Widget w, XtPointer clientData, XtPointer callData);
static void listAreaCB(Widget w, XtPointer clientData,  XtPointer callData);
void wmCloseCB(Widget w, XtPointer clientData, XtPointer callData);
int xerrMsg(const char *fmt, ...);
extern "C" int xerrorhandler(Display *dpy, XErrorEvent *event);
void xinfoMsg(char *sstring);
void xterrorhandler(char *message);
int yesNo(char *text, char *ok, char *cancel);
static void yesNoCB(Widget w, XtPointer clientData, XtPointer callData);

/* Global variables */

char colorTableMergeModeString[4][25]={
    "Match Existing Colors",
    "Replace Unused Colors",
    "Add To Existing Colors",
    "Replace Unused, Then Add"
};

CGif *gif=NULL;
int nBlocks=0;
XtAppContext appContext;
Widget appShell,mainWindow,mainMenu,mainForm,warningBox;
Widget fileSD;
Widget helpS,helpST;
Widget playRC;
Widget yesNoD;
Widget optionsS;
Widget displayModeM,backgroundM,defaultColorTableM,colorTableMergeModeM;
Window rootWindow,appShellWindow,mainWindowWindow,mainFormWindow;
Pixmap drawmap;
Atom WM_DELETE_WINDOW;
XtIntervalId playTimerID;
Boolean applyToAllImages;

GC gc;
XGCValues gcvalues;
Cursor watchCursor;
Visual *defvisual;
VisualID defvisualid;
int height,width,defplanes,defvisualclass;
int windowMessage=0;
char string[FIXED_MSG_SIZE];
char curText[FIXED_MSG_SIZE];
XmString cstring;
XColor sampleColor;
int fsdButton;

int colorTableMergeMode;
int defaultNColors,defaultNColorsBtn;

char startFileName[PATH_MAX];
int defaultNColorsArray[N_DEFAULT_COLORS]={2,4,8,16,32,64,128,256};

/**************************** main ****************************************/
int main(int argc, char **argv)
{
    Widget w,w1,wparent,wparentsav;

  /* Set the error handler */
    errHandler=xerrMsg;

  /* WIN32 initialization */
#if HCLXMINIT
#ifdef WIN32	
    HCLXmInit();
#endif
#endif

  /* Constants and defaults */
    applyToAllImages=False;
    bgMode=BACKGROUND_GIF;
    dataMode=GRAPHIC_MODE;
    colorTableMergeMode=MATCH_MODE;
    defaultNColorsBtn=DEFAULT_COLOR_N;
    defaultNColors=defaultNColorsArray[DEFAULT_COLOR_N];
    currentDrawingImage=NULL;
    currentColorTableIndex=0;
    selectedBlock=NULL;
    *startFileName='\0';
    playTimerID=(XtIntervalId)0;
    colorTableS=NULL;
    fileSD=yesNoD=optionsS=NULL;
    helpS=helpST=0;

  /* Set up top level */
    XtToolkitInitialize();
    appContext=XtCreateApplicationContext();
    XtAppSetFallbackResources(appContext,fallbackResources);

  /* Get display parameters */
    display=XtOpenDisplay(appContext,NULL,"sga","SGA",NULL,0,&argc,argv);
    if(display == NULL) {
	xerrMsg("SGA initialization: Cannot open display");
	exit(1);
    }
    screen=DefaultScreen(display);
    screenptr=ScreenOfDisplay(display,screen);
    nplanes=DisplayPlanes(display,screen);
    rootWindow=RootWindow(display,screen);
    defplanes=DefaultDepth(display,screen);
    defvisual=DefaultVisual(display,screen);
    defvisualid=XVisualIDFromVisual(defvisual);
    height=DisplayHeight(display,screen);
    width=DisplayWidth(display,screen);

  /* Parse the command line */
    if(parseCommand(argc,argv) != SGA_OK) exit(1);

  /* Set error handlers */
    XtAppSetErrorHandler(appContext,xterrorhandler);
    XtAppSetWarningHandler(appContext,xterrorhandler);
    XSetErrorHandler(xerrorhandler);

  /* Define app shell */
    nargs=0;
    XtSetArg(args[nargs],XmNtitle,"SGA"); nargs++; 
    appShell=XtAppCreateShell("sga","SGA",applicationShellWidgetClass,
      display,args,nargs);

  /* Fix up WM close */
    WM_DELETE_WINDOW=XmInternAtom(display,"WM_DELETE_WINDOW",False);
    XmAddWMProtocolCallback(appShell,WM_DELETE_WINDOW,
      quit,(XtPointer)0);

  /* Enable Editres */
    XtAddEventHandler(appShell,(EventMask)NULL,TRUE,
      (XtEventHandler)_XEditResCheckMessages,(XtPointer)NULL);

  /* Define cursors */
    watchCursor=XCreateFontCursor(display,XC_watch);

  /* Define main window */
    nargs=0;
    mainWindow=XmCreateMainWindow(appShell,"mainWindow",args,nargs);

    XtManageChild(mainWindow);

  /* Define menu bar */
    nargs=0;
    mainMenu=XmCreateMenuBar(mainWindow,"mainMenu",args,nargs);
    XtManageChild(mainMenu);
    wparent=mainMenu;

  /* Define file menu */
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w1=XmCreatePulldownMenu(mainMenu,"fileMenu",args,nargs);

    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w1); nargs++;
    w=XmCreateCascadeButton(mainMenu,"File",args,nargs);
    XtManageChild(w);

    nargs=0;
    w=XmCreatePushButton(w1,"New",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)FILE_NEW_BTN);

    nargs=0;
    w=XmCreatePushButton(w1,"Open...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)FILE_OPEN_BTN);

    nargs=0;
    w=XmCreatePushButton(w1,"Save",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)FILE_SAVE_BTN);

    nargs=0;
    w=XmCreatePushButton(w1,"Save As...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)FILE_SAVEAS_BTN);

    nargs=0;
    w=XmCreatePushButton(w1,"Quit",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)QUIT_BTN);

  /* Define edit menu */
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w1=XmCreatePulldownMenu(mainMenu,"editMenu",args,nargs);

    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w1); nargs++;
    w=XmCreateCascadeButton(mainMenu,"Edit",args,nargs);
    XtManageChild(w);

  /* Insert */
    wparentsav=wparent;
    wparent=w1;
    
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w1=XmCreatePulldownMenu(wparent,"insertMenu",args,nargs);
    
    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w1); nargs++;
    w=XmCreateCascadeButton(wparent,"Insert",args,nargs);
    XtManageChild(w);
    
    nargs=0;
    w=XmCreatePushButton(w1,"GIF...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,
      (XtPointer)INSERT_GIF_BTN);

    nargs=0;
    w=XmCreatePushButton(w1,"Comment",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,
      (XtPointer)INSERT_COMMENT_BTN);

    w1=wparent;
    wparent=wparentsav;

    nargs=0;
    w=XmCreatePushButton(w1,"Delete",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)DELETE_BTN);

    nargs=0;
    w=XmCreatePushButton(w1,"Move Up",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)MOVE_UP_BTN);

    nargs=0;
    w=XmCreatePushButton(w1,"Move Down",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)MOVE_DOWN_BTN);

  /* Define tools menu */
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w1=XmCreatePulldownMenu(wparent,"toolsMenu",args,nargs);
    
    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w1); nargs++;
    w=XmCreateCascadeButton(mainMenu,"Tools",args,nargs);
    XtManageChild(w);

  /* Display Mode */
    wparentsav=wparent;
    wparent=w1;
    
  /* Edit color table */
    nargs=0;
    w=XmCreatePushButton(wparent,"Edit Color Table...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_EDIT_BTN);
    
  /* Show unused colors */
    nargs=0;
    w=XmCreatePushButton(wparent,"Show Unused Colors...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_UNUSED_BTN);
    
  /* Trim color table */
    nargs=0;
    w=XmCreatePushButton(wparent,"Trim Color Table",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_TRIM_BTN);
    
  /* Options */
    nargs=0;
    w=XmCreatePushButton(wparent,"Options...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,
      (XtPointer)TOOLS_OPTIONS_BTN);

  /* Define help menu */
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w1=XmCreatePulldownMenu(mainMenu,"helpMenu",args,nargs);

    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w1); nargs++;
    w=XmCreateCascadeButton(mainMenu,"Help",args,nargs);
    XtManageChild(w);

    nargs=0;
    XtSetArg(args[nargs],XmNmenuHelpWidget,w); nargs++;
    XtSetValues(mainMenu,args,nargs);

    nargs=0;
    w=XmCreatePushButton(w1,"Overview",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,helpCB,(XtPointer)0);

    nargs=0;
    w=XmCreatePushButton(w1,"Version",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,helpCB,(XtPointer)1);

  /* Define form for rest of main window */
    nargs=0;
    mainForm=XmCreateForm(mainWindow,"form",args,nargs);
    XtManageChild(mainForm);

  /* Define text area */
    createMainFormAreas(mainForm);

  /* Set main window areas */
    XmMainWindowSetAreas(mainWindow,mainMenu,NULL,NULL,NULL,mainForm);

  /* Realize */
    XtRealizeWidget(appShell);

  /* Get window ids */
    appShellWindow=XtWindow(appShell);
    mainWindowWindow=XtWindow(mainWindow);
    mainFormWindow=XtWindow(mainForm);

  /* Open a file given on the command line */
    if(*startFileName) loadFile(startFileName);

#if DEBUG_SUMMARY
    if(gif) gif->printSummary();
    exit(0);
#endif
    
  /* Start */
    windowMessage=1;
    XtAppMainLoop(appContext);
    return(0);
}
/**************************** clearDisplayArea **************************/
int clearDisplayArea(const char *text)
{
    XmTextSetString(textArea,(char *)text);
    XmTextSetCursorPosition(textArea,0);
#if DEBUG_EXC
    print("clearDisplayArea: Before changeDisplayWindow\n");
#endif
    changeDisplayWindow(TEXT_MODE);
#if DEBUG_EXC
    print("clearDisplayArea: After changeDisplayWindow\n");
#endif
    return SGA_OK;
}
/**************************** clearListArea *****************************/
int clearListArea(void)
{
    
#if DEBUG_EXC
    print("clearListArea: Before XmListDeleteAllItems listArea=%x\n",listArea);
#endif
    XmListDeleteAllItems(listArea);
#if DEBUG_EXC
    print("clearListArea: After XmListDeleteAllItems listArea=%x\n",listArea);
#endif

    return SGA_OK;
}
/**************************** colorTableCB ******************************/
void colorTableCB(Widget w, XtPointer clientData, XtPointer callData)
{
    int id=(long)clientData;
    long lval;
    int val;
    char *value=NULL;

    UNREFERENCED(w);
    UNREFERENCED(callData);

  // Check if it is a color button (id=0-255)
    if(id < COLOR_TABLE_EDIT_BTN) {
	setColorTableIndex(id);
	return;
    }

    switch(id) {
    case COLOR_TABLE_EDIT_BTN:
	popupColorTable();
	break;
    case COLOR_TABLE_TRIM_BTN:
	if(gif) {
	    XDefineCursor(display,XtWindow(appShell),watchCursor);
	    XFlush(display);
	    gif->trimColorTable();
	    refreshColorTableS();
	    if(selectedBlock) selectedBlock->displayData();
	    XUndefineCursor(display,XtWindow(appShell));
	}
	break;
    case COLOR_TABLE_UNUSED_BTN:
	if(gif) {
	    XDefineCursor(display,XtWindow(appShell),watchCursor);
	    XFlush(display);
	    int nUsedColors=0;
	    int nColors=gif->getNColors();
	    for(int i=0; i < nColors; i++) {
		if(gif->isUsedColor(i)) nUsedColors++;
	    }
	    sprintf(string,"There %s %d unused of %d colors",
	      (nColors-nUsedColors == 1)?"is":"are",
	      nColors-nUsedColors,nColors);
	    xinfoMsg(string);
	    XUndefineCursor(display,XtWindow(appShell));
	}
	break;
    case COLOR_TABLE_APPLY_BTN:
	if(gif) {
	    gif->setColor(currentColorTableIndex,
	      sampleColor.red,sampleColor.green,sampleColor.blue);
	    if(selectedBlock) selectedBlock->displayData();
	    const Pixel *pixels=gif->getPixels();
	    setColorTableIndex(currentColorTableIndex);
	    nargs=0;
	    XtSetArg(args[nargs],XmNbackground,pixels[currentColorTableIndex]);
	    nargs++;
	    XtSetValues(colorButtons[currentColorTableIndex],args,nargs);
	}
	break;
    case COLOR_TABLE_CLOSE_BTN:
	XtPopdown(colorTableS);
	break;
    case COLOR_TABLE_INDEX_ENTRY:
	value=XmTextGetString(colorTableIndexEntry);
	lval=atol(value);
	if(lval < 0) lval=0;
	if(lval > 0xff) lval=0xff;
	setColorTableIndex(lval);
	break;
    case COLOR_TABLE_RED_ENTRY:
	value=XmTextGetString(colorTableRedEntry);
	lval=atol(value);
	if(lval < 0) lval=0;
	if(lval > 0xff) lval=0xff;
	updateColorTableSample(DO_RED,lval);
	break;
    case COLOR_TABLE_GREEN_ENTRY:
	value=XmTextGetString(colorTableGreenEntry);
	lval=atol(value);
	if(lval < 0) lval=0;
	if(lval > 0xff) lval=0xff;
	updateColorTableSample(DO_GREEN,lval);
	break;
    case COLOR_TABLE_BLUE_ENTRY:
	value=XmTextGetString(colorTableBlueEntry);
	lval=atol(value);
	if(lval < 0) lval=0;
	if(lval > 0xff) lval=0xff;
	updateColorTableSample(DO_BLUE,lval);
	break;
    case COLOR_TABLE_RED_SCALE:
	XmScaleGetValue(colorTableRedScale,&val);
	updateColorTableSample(DO_RED,val);
	break;
    case COLOR_TABLE_GREEN_SCALE:
	XmScaleGetValue(colorTableGreenScale,&val);
	updateColorTableSample(DO_GREEN,val);
	break;
    case COLOR_TABLE_BLUE_SCALE:
	XmScaleGetValue(colorTableBlueScale,&val);
	updateColorTableSample(DO_BLUE,val);
	break;
    }
}
/**************************** optCB**************************************/
void optCB(Widget w, XtPointer clientData, XtPointer callData)
{
    int id=(long)clientData;
    XtPointer userData;
    CImage *image=NULL;
    CGlobal *global=NULL;
    CComment *comment=NULL;
    Boolean state;
    char *value=NULL;

    UNREFERENCED(callData);

  // Get the userData which is a pointer to the class instance
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,&userData); nargs++;
    XtGetValues(w,args,nargs);

    switch(id) {
    case APPLY_ALL_BTN:
	state=XmToggleButtonGetState(w);
	applyToAllImages=state;
	break;
    case TRANSPARENT_COLOR_BTN:
	image=(CImage *)userData;
	state=XmToggleButtonGetState(w);
	if(applyToAllImages) {
	    gif->setTransparentColorFlag(state);
	} else {
	    image->setTransparentColorFlag(state);
	}
	image->displayData();
	break;
    case INTERLACE_BTN:
	image=(CImage *)userData;
	state=XmToggleButtonGetState(w);
	image->setInterlace(state);
	if(applyToAllImages) {
	    gif->setTransparentColorFlag(state);
	} else {
	    image->setTransparentColorFlag(state);
	}
	break;
    case LOCAL_COLOR_TABLE_BTN:
	image=(CImage *)userData;
	state=XmToggleButtonGetState(w);
	if(applyToAllImages) {
	    gif->setLocalColorTableFlag(state);
	} else {
	    image->setLocalColorTableFlag(state);
	}
	break;
    case TRANSPARENT_INDEX_ENTRY:
	image=(CImage *)userData;
	value=XmTextFieldGetString(w);
	if(applyToAllImages) {
	    gif->setTransparentIndex(value);
	} else {
	    image->setTransparentIndex(value);
	}
	XtFree(value);
	image->displayData();
	break;
    case DELAY_TIME_ENTRY:
	image=(CImage *)userData;
	value=XmTextFieldGetString(w);
	if(applyToAllImages) {
	    gif->setDelayTime(value);
	} else {
	    image->setDelayTime(value);
	}
	XtFree(value);
	break;
    case LEFT_OFFSET_ENTRY:
	image=(CImage *)userData;
	value=XmTextFieldGetString(w);
	if(applyToAllImages) {
	    gif->setLeftOffset(value);
	} else {
	    image->setLeftOffset(value);
	}
	XtFree(value);
	image->displayData();
	break;
    case TOP_OFFSET_ENTRY:
	image=(CImage *)userData;
	value=XmTextFieldGetString(w);
	if(applyToAllImages) {
	    gif->setTopOffset(value);
	} else {
	    image->setTopOffset(value);
	}
	XtFree(value);
	image->displayData();
	break;
    case BG_COLOR_INDEX_ENTRY:
	global=(CGlobal *)userData;
	value=XmTextFieldGetString(w);
	global->setBackgroundColorIndex(value);
	XtFree(value);
	break;
    case LOGICAL_SCREEN_WIDTH_ENTRY:
	global=(CGlobal *)userData;
	value=XmTextFieldGetString(w);
	global->setLogicalScreenWidth(value);
	XtFree(value);
	break;
    case LOGICAL_SCREEN_HEIGHT_ENTRY:
	global=(CGlobal *)userData;
	value=XmTextFieldGetString(w);
	global->setLogicalScreenHeight(value);
	XtFree(value);
	break;
    case COMMENT_APPLY_BTN:
	comment=(CComment *)userData;
	value=XmTextGetString(commentText);
	comment->setComment(value);
	comment->displayData();
	XtFree(value);
	break;
    }
}
/**************************** optCB**************************************/
void disposalMethodCB(Widget w, XtPointer clientData, XtPointer callData)
{
    int method=(long)clientData;
    XtPointer userData;
    CImage *image=NULL;
    
    UNREFERENCED(w);
    UNREFERENCED(callData);

  // Get the userData which is a pointer to the class instance
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,&userData); nargs++;
    XtGetValues(disposalMethodMenu,args,nargs);
    image=(CImage *)userData;

  // Set the disposal method
    if(image) image->setDisposalMethod(method);

//      print("disposalMethodCB: w=%x parent=%x disposalMethodMenu=%x "
//        "method=%d userData=%x\n",
//        w,XtParent(w),disposalMethodMenu,method,userData);
}
/**************************** createMainFormAreas *************************/
void disposalMethodSet(int buttonId)
{
    WidgetList buttons;
    Cardinal numButtons;
    Widget subMenu;

    XtVaGetValues(disposalMethodMenu,XmNsubMenuId,&subMenu,NULL);
    if (subMenu != NULL) {
	XtVaGetValues(subMenu, XmNchildren,&buttons, XmNnumChildren,
	  &numButtons, NULL);
	if (buttonId < (int)numButtons && buttonId >= 0) {
	    XtVaSetValues(disposalMethodMenu,XmNmenuHistory,buttons[buttonId],
	      NULL);
	}
    }
}
/**************************** createMainFormAreas *************************/
void createMainFormAreas(Widget wparent)
{
    Widget w;
    Widget pane;
    Widget actionArea;
    Widget controlArea,controlRC;
    Widget optionArea;

  /* Create paned window */
    pane=XtVaCreateWidget("panel",xmPanedWindowWidgetClass,wparent,
      XmNtopAttachment,XmATTACH_FORM,
      XmNbottomAttachment,XmATTACH_FORM,
      XmNleftAttachment,XmATTACH_FORM,
      XmNrightAttachment,XmATTACH_FORM,
      XmNsashWidth,1,     /* Sashes won't show */
      XmNsashHeight,1,
      NULL);
    
  /* Create action area */
    actionArea=XtVaCreateWidget("actionArea",
      xmFormWidgetClass,pane,
      XmNshadowThickness,0,
      XmNskipAdjust,False,
      NULL);

  /* Create scrolled list */
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNselectionPolicy,XmSINGLE_SELECT); nargs++;
    XtSetArg(args[nargs],XmNwidth,LIST_WIDTH); nargs++;
    XtSetArg(args[nargs],XmNvisibleItemCount,LIST_VISIBLE); nargs++;
    XtSetArg(args[nargs],XmNlistSizePolicy,XmCONSTANT); nargs++;
    XtSetArg(args[nargs],XmNscrollBarDisplayPolicy,XmSTATIC); nargs++;
    XtSetArg(args[nargs],XmNtraversalOn,False); nargs++;
    listArea=XmCreateScrolledList(actionArea,"listArea",args,nargs);
    XtManageChild(listArea);
  //    XtAddCallback(listArea,XmNdefaultActionCallback,listAreaCB,NULL);
    XtAddCallback(listArea,XmNsingleSelectionCallback,listAreaCB,NULL);

  /* Create text scrolled window */
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_WIDGET); nargs++;
    XtSetArg(args[nargs],XmNleftWidget,listArea); nargs++;
    XtSetArg(args[nargs],XmNleftOffset,5); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNheight,GRAPH_HEIGHT); nargs++;
    XtSetArg(args[nargs],XmNwidth,GRAPH_WIDTH); nargs++;
  /* XtSetArg(args[nargs],XmNscrollBarDisplayPolicy,XmSTATIC); nargs++; */
    textSW=XmCreateScrolledWindow(actionArea,"textSW",args,nargs);
    XtManageChild(textSW);

  /* Create text area */
    nargs=0;
    XtSetArg(args[nargs],XmNrows,N_TEXT_ROWS); nargs++;
    XtSetArg(args[nargs],XmNcolumns,N_TEXT_COLS); nargs++;
    XtSetArg(args[nargs],XmNeditMode,XmMULTI_LINE_EDIT); nargs++;
    XtSetArg(args[nargs],XmNeditable,False); nargs++;
    XtSetArg(args[nargs],XmNvalue,WELCOME); nargs++;
    XtSetArg(args[nargs],XmNtraversalOn,False); nargs++;
#if 1
    XtSetArg(args[nargs],XmNbackground,WhitePixel(display,screen)); nargs++;  /* Exceed bug */
#endif
    textArea=XmCreateText(textSW,"textArea",args,nargs);
    XtManageChild(textArea);

  /* Create drawing scrolled window */
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_WIDGET); nargs++;
    XtSetArg(args[nargs],XmNleftWidget,listArea); nargs++;
    XtSetArg(args[nargs],XmNleftOffset,5); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNheight,GRAPH_HEIGHT); nargs++;
    XtSetArg(args[nargs],XmNwidth,GRAPH_WIDTH); nargs++;
    XtSetArg(args[nargs],XmNscrollBarDisplayPolicy,XmSTATIC); nargs++;
    XtSetArg(args[nargs],XmNscrollingPolicy,XmAUTOMATIC); nargs++;
    drawingSW=XmCreateScrolledWindow(actionArea,"drawingSW",args,nargs);
  /* (Don't manage now) */

  /* Create drawing area */
    nargs=0;
    XtSetArg(args[nargs],XmNheight,GRAPH_HEIGHT); nargs++;
    XtSetArg(args[nargs],XmNwidth,GRAPH_WIDTH); nargs++;
    XtSetArg(args[nargs],XmNtraversalOn,False); nargs++;
    drawingArea=XmCreateDrawingArea(drawingSW,"drawingArea",args,nargs);
    XtManageChild(drawingArea);
    XtAddCallback(drawingArea,XmNexposeCallback,drawingAreaCB,(XtPointer)NULL);
    XtAddCallback(drawingArea,XmNresizeCallback,drawingAreaCB,(XtPointer)NULL);

  /* Create control area */
    nargs=0;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNskipAdjust,True); nargs++;
    controlArea=XmCreateDrawingArea(pane,"controlArea",args,nargs);

  /* Create control row column */
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    controlRC=XmCreateRowColumn(controlArea,"controlRC",args,nargs);

 /* Define play control */
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    playRC=XmCreateRowColumn(controlRC,"playRC",args,nargs);
    XtManageChild(playRC);

    nargs=0;
    w=XmCreatePushButton(playRC,"Play",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)PLAY_BTN);

    nargs=0;
    w=XmCreatePushButton(playRC,"Stop",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)STOP_BTN);

    nargs=0;
    w=XmCreatePushButton(playRC,"First",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)FIRST_BTN);

    nargs=0;
    w=XmCreatePushButton(playRC,"Next",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)NEXT_BTN);

    nargs=0;
    w=XmCreatePushButton(playRC,"Prev",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)PREV_BTN);

    nargs=0;
    w=XmCreatePushButton(playRC,"Last",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)LAST_BTN);

  /* Create option area */
    nargs=0;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNskipAdjust,True); nargs++;
    optionArea=XmCreateDrawingArea(pane,"optionArea",args,nargs);

  /* Create option areas */
    defaultOptionArea=createDefaultOptionArea(optionArea);
    globalOptionArea=createGlobalOptionArea(optionArea);
    imageOptionArea=createImageOptionArea(optionArea);
    commentOptionArea=createCommentOptionArea(optionArea);

  /* Manage the managers last */
    XtManageChild(actionArea);
    XtManageChild(controlRC);
    XtManageChild(controlArea);
    XtManageChild(defaultOptionArea);
    XtManageChild(optionArea);
    XtManageChild(pane);

#if 0
  /* Initialize */
    if(errMsgS) {
	long now; 
	struct tm *tblock;
	char timeStampStr[TIME_STRING_MAX];
	XmTextPosition curpos=0;

	time(&now);
	tblock=localtime(&now);
	strftime(timeStampStr,TIME_STRING_MAX,
	  "Message Log started at "STRFTIME_FORMAT"\n",tblock);
	timeStampStr[TIME_STRING_MAX-1]='0';
	XmTextInsert(errMsgText, curpos, timeStampStr);
	curpos+=strlen(timeStampStr);
	XmTextSetInsertionPosition(errMsgText, curpos);
    }
#endif
}
/**************************** displayMsg **********************************/
void displayMsg(char *fmt, ...)
{
    va_list vargs;
    static char lstring[1024];
    
    va_start(vargs,fmt);
    (void)vsprintf(lstring,fmt,vargs);
    va_end(vargs);
    
    if(lstring[0] != '\0') {
	XmTextSetString(textArea,lstring);
    } else {
	XmTextSetString(textArea,"");
    }
}
/**************************** helpCB **************************************/
void helpCB(Widget w, XtPointer clientData, XtPointer callData)
{
    Widget w1,wparent,form;
    int i,n,nHelpStrings;
    n=(long)clientData;

    UNREFERENCED(w);
    UNREFERENCED(callData);

    switch(n) {
    case 0:
	if(!helpS) {
	    nargs=0;
	    XtSetArg(args[nargs],XmNtitle,"Help"); nargs++;
	    XtSetArg(args[nargs],XmNdeleteResponse, XmDO_NOTHING); nargs++;
	    helpS=XtCreatePopupShell("helpS",topLevelShellWidgetClass,
	      appShell,args,nargs);
	    XmAddWMProtocolCallback(helpS,WM_DELETE_WINDOW,
	      wmCloseCB,NULL);
	    
	    nargs=0;
	    form=XmCreateForm(helpS,"optPopup",args,nargs);
	    
	    nargs=0;
	    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
	    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
	    w1=XmCreatePushButton(form,"Remove",args,nargs);
	    XtManageChild(w1);
	    XtAddCallback(w1,XmNactivateCallback,helpCB,(XtPointer)2);
	    wparent=w1;
	    
	    nargs=0;
	    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
	    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_WIDGET); nargs++;
	    XtSetArg(args[nargs],XmNleftWidget,w1); nargs++;
	    w1=XmCreatePushButton(form,"Select All",args,nargs);
	    XtManageChild(w1);
	    XtAddCallback(w1,XmNactivateCallback,helpCB,(XtPointer)3);
	    
	    nargs=0;
	    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_WIDGET); nargs++;
	    XtSetArg(args[nargs],XmNtopWidget,wparent); nargs++;
	    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
	    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
	    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
	    XtSetArg(args[nargs],XmNeditMode,XmMULTI_LINE_EDIT); nargs++;
	    XtSetArg(args[nargs],XmNeditable,False); nargs++;
	    XtSetArg(args[nargs],XmNcursorPositionVisible,False); nargs++;
	    XtSetArg(args[nargs],XmNrows,25); nargs++;
	    XtSetArg(args[nargs],XmNcolumns,80); nargs++;
	    w1=XmCreateScrolledText(form,"scrolledHelp",args,nargs);
	    helpST=w1;

	  // The help text must be broken down into string literals of
	  // length less than 2048 for WIN32.  Loop and set these strings
	  // into the text field.
	    nHelpStrings=sizeof(helptext)/sizeof(char *);
	    XmTextSetString(helpST,helptext[0]);
	    for(i=1; i < nHelpStrings; i++) {
		XmTextInsert(helpST,XmTextGetLastPosition(helpST),helptext[i]);
	    }
	    XtManageChild(helpST);
	    
	    XtManageChild(form);
	}

	if(helpS) XtPopup(helpS,XtGrabNone);
	break;
    case 1:
	sprintf(string,"%s\nCompilation Date: %s %s",SGA_VERSION_STRING,
	  __DATE__,__TIME__);
	xinfoMsg(string);
	break;
    case 2:
	if(helpS) XtPopdown(helpS);
	break;
    case 3:
	if(helpS) {
	    XmTextSetSelection(helpST,0,XmTextGetLastPosition(helpST),CurrentTime);
	    XmTextShowPosition(helpST,0);
	}
	break;
    }
}
/**************************** getLogo *************************************/
Pixmap getLogo(Pixel bg, char **xpm)
{
    static CXpm *logo=NULL;
    Pixmap pixmap=(Pixmap)NULL;
    int error=0;

  // Clean up the old file
    if(logo) {
	delete logo;
	logo=NULL;
    }

  // Allocate a new CXpm
    logo=new CXpm(display,xpm,bg,xerrMsg);
    if(!logo) {
	xerrMsg("Cannot create class for logo\n");
	error=1;
	goto CLEANUP;
    }
    pixmap=logo->getPixmap();

  // Clean up
  CLEANUP:
    if(error) {
	if(logo) {
	    delete logo;
	    logo=NULL;
	}
	return (Pixmap)NULL;
    }

    return pixmap;
}
/**************************** mainCB **************************************/
void mainCB(Widget w, XtPointer clientData, XtPointer callData)
{
    int id=(long)clientData;
    CBlock *block;
    int pos;

    switch(id) {
    case QUIT_BTN:
	quit(w,clientData,callData);
	break;
    case FILE_NEW_BTN:
	loadFile(NULL);
	break;
    case FILE_OPEN_BTN:
	fsdButton=FILE_OPEN_BTN;
	executeFSD();
	break;
    case FILE_SAVE_BTN:
	if(gif) {
	    const char *fileName=gif->getLongName();
	    if(!fileName || !*fileName) {
	      // Is new
		fsdButton=FILE_SAVEAS_BTN;
		executeFSD();
	    } else {
		XDefineCursor(display,XtWindow(appShell),watchCursor);
		XFlush(display);
		gif->writeFile(gif->getLongName());
		XUndefineCursor(display,XtWindow(appShell));
	    }
	}
	break;
    case FILE_SAVEAS_BTN:
	fsdButton=FILE_SAVEAS_BTN;
	executeFSD();
	break;
    case INSERT_GIF_BTN:
	if(!gif || !selectedBlock) {
	    XBell(display,50);
	    break;
	}
	fsdButton=INSERT_GIF_BTN;
	executeFSD();
	break;
    case INSERT_COMMENT_BTN:
	stopPlay();
	if(!gif || !selectedBlock) {
	    XBell(display,50);
	    break;
	}
	block=gif->addComment(selectedBlock);
	if(block) {
	    gif->markEdited();
	    gif->updateListArea();
	    selectedBlock=(CBlock *)block;
	    pos=gif->getPos(block);
	    if(pos) XmListSelectPos(listArea,pos,True);
	    block->displayData();
	} else {
	    XBell(display,50);
	}
	break;
    case PLAY_BTN:
	startPlay();
	break;
    case STOP_BTN:
	stopPlay();
	break;
    case FIRST_BTN:
	stopPlay();
	if(!gif) {
	    XBell(display,50);
	    break;
	}
	block=gif->getFirstImage();
	if(block) {
	    selectedBlock=(CBlock *)block;
	    pos=gif->getPos(block);
	    if(pos) XmListSelectPos(listArea,pos,True);
	    block->displayData();
	} else {
	    xerrMsg("Cannot find first image");
	}
	break;
    case NEXT_BTN :
	stopPlay();
	if(!gif) {
	    XBell(display,50);
	    break;
	}
	if(selectedBlock) {
	    block=gif->getNextImage(selectedBlock);
	} else {
	    block=gif->getLastImage();
	}
	if(block) {
	    selectedBlock=(CBlock *)block;
	    pos=gif->getPos(block);
	    if(pos) XmListSelectPos(listArea,pos,True);
	    block->displayData();
	} else {
	    XBell(display,50);
	}
	break;
    case PREV_BTN:
	stopPlay();
	if(!gif) {
	    XBell(display,50);
	    break;
	}
	if(selectedBlock) {
	    block=gif->getPrevImage(selectedBlock);
	} else {
	    block=gif->getFirstImage();
	}
	if(block) {
	    selectedBlock=(CBlock *)block;
	    pos=gif->getPos(block);
	    if(pos) XmListSelectPos(listArea,pos,True);
	    block->displayData();
	} else {
	    XBell(display,50);
	}
	break;
    case LAST_BTN :
	stopPlay();
	if(!gif) {
	    XBell(display,50);
	    break;
	}
	block=gif->getLastImage();
	if(block) {
	    selectedBlock=(CBlock *)block;
	    pos=gif->getPos(block);
	    if(pos) XmListSelectPos(listArea,pos,True);
	    block->displayData();
	} else {
	    xerrMsg("Cannot find last image");
	}
	break;
    case DELETE_BTN:
	stopPlay();
	if(!gif || !selectedBlock) {
	    XBell(display,50);
	    break;
	}
	block=gif->deleteBlock(selectedBlock);
	if(block) {
	    gif->markEdited();
	    gif->updateListArea();
	    selectedBlock=(CBlock *)block;
	    pos=gif->getPos(block);
	    if(pos) XmListSelectPos(listArea,pos,True);
	    block->displayData();
	} else {
	    XBell(display,50);
	}
	break;
    case MOVE_UP_BTN:
	stopPlay();
	if(!gif || !selectedBlock) {
	    XBell(display,50);
	    break;
	}
	block=gif->moveBlockUp(selectedBlock);
	if(block) {
	    gif->markEdited();
	    gif->updateListArea();
	    selectedBlock=(CBlock *)block;
	    pos=gif->getPos(block);
	    if(pos) XmListSelectPos(listArea,pos,True);
	    block->displayData();
	} else {
	    XBell(display,50);
	}
	break;
    case MOVE_DOWN_BTN:
	stopPlay();
	if(!gif || !selectedBlock) {
	    XBell(display,50);
	    break;
	}
	block=gif->moveBlockDown(selectedBlock);
	if(block) {
	    gif->markEdited();
	    gif->updateListArea();
	    selectedBlock=(CBlock *)block;
	    pos=gif->getPos(block);
	    if(pos) XmListSelectPos(listArea,pos,True);
	    block->displayData();
	} else {
	    XBell(display,50);
	}
	break;
    case TOOLS_OPTIONS_BTN:
	popupOptionsDialog();
	break;
    case OPTIONS_CANCEL_BTN:
	XtPopdown(optionsS);
	break;
    }
}
/**************************** optionsCB ***********************************/
static void optionsCB(Widget w, XtPointer clientData, XtPointer callData)
{
    XtPointer userData;
    int button=(long)clientData;

    UNREFERENCED(callData);

  // Get the user data
    XtVaGetValues(XtParent(w), XmNuserData, &userData, NULL);

    switch((long)userData) {
    case OPTIONS_DISPLAY_MODE_BTN:
	if(button >= 0 && button < 2) {
	    dataMode=button;
	    if(gif) gif->setDataMode(dataMode);
	    if(selectedBlock) selectedBlock->displayData();
	} else {
	    xerrMsg("optionsCB:  Invalid button for Display Mode: %d\n",
	      button);
	}
	break;
    case OPTIONS_BACKGROUND_BTN:
	if(button >= 0 && button < 3) {
	    bgMode=button;
	    if(gif) gif->setBgMode(bgMode);
	    if(currentDrawingImage) currentDrawingImage->resizeCB();
	} else {
	    xerrMsg("optionsCB:  Invalid button for Background: %d\n",
	      button);
	}
	break;
    case OPTIONS_DEF_COLORTABLE_BTN:
	if(button >= 0 && button < N_DEFAULT_COLORS) {
	    defaultNColorsBtn=button;
	    defaultNColors=defaultNColorsArray[defaultNColorsBtn];
	} else if(button == N_DEFAULT_COLORS) {
	    xinfoMsg("To use a custom color table, load a GIF with the desired\n"
	      " color table rather than using the File New button.");
	    optionMenuSet(defaultColorTableM,defaultNColorsBtn);
	} else {
	    xerrMsg("optionsCB:  Invalid button for Default Color Table: %d\n",
	      button);
	}
	break;
    case OPTIONS_MERGE_MODE_BTN:
	if(button >= 0 && button < 4) {
	    colorTableMergeMode=button;
	} else {
	    xerrMsg("optionsCB:  Invalid button for Merge Mode: %d\n",
	      button);
	}
	break;
    }
}
/**************************** optionsMenuSet ******************************/
static void optionMenuSet(Widget menu, int buttonId)
{
    WidgetList buttons;
    Cardinal numButtons;
    Widget subMenu;

    XtVaGetValues(menu,XmNsubMenuId,&subMenu,NULL);
    if(subMenu != NULL) {
	XtVaGetValues(subMenu,XmNchildren,&buttons,XmNnumChildren,
	  &numButtons,NULL);
	if(buttonId < (int)numButtons && buttonId >= 0) {
	    XtVaSetValues(menu,XmNmenuHistory,buttons[buttonId],NULL);
	}
    }
}
/**************************** createOptionsDialog**************************/
static int createOptionsDialog(void)
{
    Widget form,actionArea;
    Widget w,wparent;
    XmString label,opt1,opt2,opt3,opt4,opt5,opt6,opt7,opt8,opt9;

    if(optionsS) return SGA_OK;

    nargs=0;
    XtSetArg(args[nargs],XmNtitle,"Options"); nargs++;
    XtSetArg(args[nargs],XmNdeleteResponse,XmDO_NOTHING); nargs++;
    optionsS=XtCreatePopupShell("optionsS",topLevelShellWidgetClass,
      appShell,args,nargs);
    XmAddWMProtocolCallback(optionsS,WM_DELETE_WINDOW,
      wmCloseCB,NULL);
    
    nargs=0;
    form=wparent=XmCreateForm(optionsS,"optPopup",args,nargs);
    
  // Display Mode
    label=XmStringCreateLocalized("Display Mode");
    opt1=XmStringCreateLocalized("Data");
    opt2=XmStringCreateLocalized("Image");
    w=XmVaCreateSimpleOptionMenu(wparent,"displayModeM",
      label,'\0',0,optionsCB,
      XmVaPUSHBUTTON,opt1,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt2,'\0',NULL,NULL,
      XmNuserData,(XtPointer)(OPTIONS_DISPLAY_MODE_BTN),
      XmNtopAttachment,XmATTACH_FORM,
      XmNleftAttachment,XmATTACH_FORM,
      XmNrightAttachment,XmATTACH_FORM,
      NULL);
    XtManageChild(w);
    XmStringFree(label);
    XmStringFree(opt1);
    XmStringFree(opt2);
    displayModeM=w;

  // Background
    label=XmStringCreateLocalized("Background");
    opt1=XmStringCreateLocalized("GIF Background");
    opt2=XmStringCreateLocalized("White");
    opt3=XmStringCreateLocalized("Black");
    w=XmVaCreateSimpleOptionMenu(wparent,"backgroundM",
      label,'\0',0,optionsCB,
      XmVaPUSHBUTTON,opt1,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt2,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt3,'\0',NULL,NULL,
      XmNuserData,(XtPointer)(OPTIONS_BACKGROUND_BTN),
      XmNtopAttachment,XmATTACH_WIDGET,
      XmNtopWidget,w,
      XmNleftAttachment,XmATTACH_FORM,
      XmNrightAttachment,XmATTACH_FORM,
      NULL);
    XtManageChild(w);
    XmStringFree(label);
    XmStringFree(opt1);
    XmStringFree(opt2);
    XmStringFree(opt3);
    backgroundM=w;

  // Default ColorTable
    label=XmStringCreateLocalized("Default Color Table");
    opt1=XmStringCreateLocalized("Monochrome");
    opt2=XmStringCreateLocalized("4 Colors");
    opt3=XmStringCreateLocalized("8 Colors");
    opt4=XmStringCreateLocalized("16 Colors");
    opt5=XmStringCreateLocalized("32 Colors");
    opt6=XmStringCreateLocalized("64 Colors");
    opt7=XmStringCreateLocalized("128 Colors");
    opt8=XmStringCreateLocalized("256 Colors");
    opt9=XmStringCreateLocalized("Custom");
    w=XmVaCreateSimpleOptionMenu(wparent,"defaultColorTableM",
      label,'\0',0,optionsCB,
      XmVaPUSHBUTTON,opt1,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt2,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt3,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt4,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt5,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt6,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt7,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt8,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt9,'\0',NULL,NULL,
      XmNuserData,(XtPointer)(OPTIONS_DEF_COLORTABLE_BTN),
      XmNtopAttachment,XmATTACH_WIDGET,
      XmNtopWidget,w,
      XmNleftAttachment,XmATTACH_FORM,
      XmNrightAttachment,XmATTACH_FORM,
      NULL);
    XtManageChild(w);
    XmStringFree(label);
    XmStringFree(opt1);
    XmStringFree(opt2);
    XmStringFree(opt3);
    XmStringFree(opt4);
    XmStringFree(opt5);
    XmStringFree(opt6);
    XmStringFree(opt7);
    XmStringFree(opt8);
    XmStringFree(opt9);
    defaultColorTableM=w;

  // Color Table Merge Mode
    label=XmStringCreateLocalized("Color Table Merge Mode");
    opt1=XmStringCreateLocalized(colorTableMergeModeString[0]);
    opt2=XmStringCreateLocalized(colorTableMergeModeString[1]);
    opt3=XmStringCreateLocalized(colorTableMergeModeString[2]);
    opt4=XmStringCreateLocalized(colorTableMergeModeString[3]);
    w=XmVaCreateSimpleOptionMenu(wparent,"defaultColorTableM",
      label,'\0',0,optionsCB,
      XmVaPUSHBUTTON,opt1,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt2,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt3,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt4,'\0',NULL,NULL,
      XmNuserData,(XtPointer)(OPTIONS_MERGE_MODE_BTN),
      XmNtopAttachment,XmATTACH_WIDGET,
      XmNtopWidget,w,
      XmNleftAttachment,XmATTACH_FORM,
      XmNrightAttachment,XmATTACH_FORM,
      NULL);
    XtManageChild(w);
    XmStringFree(label);
    XmStringFree(opt1);
    XmStringFree(opt2);
    XmStringFree(opt3);
    XmStringFree(opt4);
    colorTableMergeModeM=w;

  // Create action area
    wparent=form;
    w=XtVaCreateWidget("optionsActionArea",
      xmFormWidgetClass,wparent,
      XmNshadowThickness,0,
      XmNfractionBase,5,
      XmNtopAttachment,XmATTACH_WIDGET,
      XmNtopWidget,w,
      XmNleftAttachment,XmATTACH_FORM,
      XmNrightAttachment,XmATTACH_FORM,
      XmNbottomAttachment,XmATTACH_FORM,
      NULL);
    actionArea=wparent=w;

    w=XtVaCreateManagedWidget("Cancel",
      xmPushButtonWidgetClass,wparent,
      XmNtopAttachment,XmATTACH_FORM,
      XmNbottomAttachment,XmATTACH_FORM,
      XmNleftAttachment,XmATTACH_POSITION,
      XmNleftPosition,1,
      XmNrightAttachment,XmATTACH_POSITION,
      XmNrightPosition,2,
      NULL);
    XtAddCallback(w,XmNactivateCallback,
      mainCB,(XtPointer)OPTIONS_CANCEL_BTN);

    w=XtVaCreateManagedWidget("Help",
      xmPushButtonWidgetClass,wparent,
      XmNtopAttachment,XmATTACH_FORM,
      XmNbottomAttachment,XmATTACH_FORM,
      XmNleftAttachment,XmATTACH_POSITION,
      XmNleftPosition,3,
      XmNrightAttachment,XmATTACH_POSITION,
      XmNrightPosition,4,
      NULL);
    XtAddCallback(w,XmNactivateCallback,helpCB,(XtPointer)0);

  // Manage the managers
    XtManageChild(actionArea);
    XtManageChild(form);
    return SGA_OK;
}
/**************************** popupOptionsDialog***************************/
static int popupOptionsDialog(void)
{
    int status;

  // Create the dialog if it has not been created
    if(!optionsS) {
	status=createOptionsDialog();
	if(status != SGA_OK || optionsS == NULL) return status;
    }

  // Set the dialog
    optionMenuSet(displayModeM,dataMode);
    optionMenuSet(backgroundM,bgMode);
    optionMenuSet(defaultColorTableM,defaultNColorsBtn);
    optionMenuSet(colorTableMergeModeM,colorTableMergeMode);

    XtPopup(optionsS,XtGrabNone);
    
    return SGA_OK;
}
/**************************** parseCommand ********************************/
int parseCommand(int argc, char **argv)
{
    int startFileSpecified=0;

    for(int i=1; i < argc; i++) {
	if (argv[i][0] == '-') {
	    switch(argv[i][1]) {
#if 0
	    case 's':
		doServer=1;
		serverName=argv[++i];
		break;
#endif
	    case 'h':
		usage();
		return SGA_OK;
	    default:
		fprintf(stderr,"\n\nInvalid option: %s\n",argv[i]);
		usage();
		return SGA_ERROR;
	    }
	} else {
	    if(!startFileSpecified) {
		strcpy(startFileName,argv[i]);
		startFileSpecified=1;
	    } else {
		fprintf(stderr,"\n\nInvalid option: %s\n",argv[i]);
		usage();
		return SGA_ERROR;
	    }
	}
    }
    return SGA_OK;
}
/**************************** print ***************************************/
void print(const char *fmt, ...)
{
    va_list vargs;
    static char lstring[1024];  /* DANGER: Fixed buffer size */
    
    va_start(vargs,fmt);
    vsprintf(lstring,fmt,vargs);
    va_end(vargs);
    
    if(lstring[0] != '\0') {
#ifdef WIN32
	lprintf("%s",lstring);
#else
	printf("%s",lstring);
#endif
    }
}
/**************************** printMsg ************************************/
void printMsg(void)
{
#if 0
    char *text;
    int first=1;
    static  char *hostName;

  // Get hostname
    if(first) {
	struct utsname ubuf;
	
	if(uname(&ubuf) < 0)
	  hostName=strDup("Unknown");
	else
	  hostName=strDup(ubuf.nodename);
    }

    text=XmTextGetString(textArea);

    print("\n%s Host=%s Server=%s\n",
      timeStamp(),hostName,serverName?serverName:"Unknown");
    if(text) {
	printf(text);
	XtFree(text);
    }
#endif
}
/**************************** quit ****************************************/
void quit(Widget w, XtPointer clientData, XtPointer callData)
{
    UNREFERENCED(w);
    UNREFERENCED(clientData);
    UNREFERENCED(callData);

    XtDestroyWidget(appShell);
    if(gif) delete gif;
    gif=NULL;
    exit(0);
}
/**************************** refreshColorTableS **************************/
void refreshColorTableS(void)
{
    if(colorTableS && XtIsRealized(colorTableS)) {
	XWindowAttributes attr;

	XGetWindowAttributes(display,XtWindow(colorTableS),&attr);
	if(attr.map_state != IsUnmapped) popupColorTable();
    }
}
/**************************** setWindowTitle ******************************/
int setWindowTitle(char * fileName)
{
    char *ptr;
    char dirName[PATH_MAX];
    char name[TITLE_MAX];
    int dirLength=0;

    if(fileName) {
	strcpy(dirName,fileName);
	ptr=strrchr(dirName, DIR_DELIMITER_CHAR);
	if(ptr) {
	    *(++ptr)='\0';
	    dirLength=strlen(dirName);
	} else {
	    dirLength=0;
	}
	strcpy(name,"SGA: ");
	strncpy(name+5,fileName+dirLength,TITLE_MAX-5);
	name[TITLE_MAX-1]='\0';
	nargs=0;
    } else {
	strcpy(name,"SGA: New");
    }
    nargs=0;
    XtSetArg(args[nargs],XmNtitle,name); nargs++;
    XtSetValues(appShell,args,nargs);

    return dirLength;
}
/**************************** playTimerCB *********************************/
void playTimerCB(XtPointer clientData, XtIntervalId *id)
{
    CBlock *block;
    int delayTime;

    UNREFERENCED(clientData);
    UNREFERENCED(id);

    if(!gif) return;

  // Use the first block if nothing is selected
    if(!selectedBlock) {
	block=gif->getFirstImage();
	if(!block) {
	    xerrMsg("playTimerCB: Cannot find first block");
	    return;
	}
    } else {
	block=gif->getNextImage(selectedBlock);
	if(!block) {
	    block=gif->getFirstImage();
	    if(!block) {
		xerrMsg("playTimerCB: Cannot find a block");
		return;
	    }
	}
    }

  // Display the block
    selectedBlock=(CBlock *)block;
    int pos=gif->getPos(block);
    if(pos) XmListSelectPos(listArea,pos,True);
    block->displayData();

  // Redisplay in case something happened
    XmUpdateDisplay(appShell);

  // Reinstall the timer if there is more than one image
  // (GIF delay time is 1/100 sec, Xt timer is 1/1000)
    if(gif->getNImages() > 1) {
	delayTime=10*((CImage *)block)->getDelayTime();
	playTimerID=XtAppAddTimeOut(appContext,delayTime,playTimerCB,
	  (XtPointer)NULL);
    }
}
/**************************** timestamp ***********************************/
char *timeStamp(void)
  // Gets current time and puts it in a static array
  // The calling program should copy it to a safe place
  //   e.g. strcpy(savetime,timestamp());
{
	static char timeStampStr[16];
	time_t now;
	struct tm *tblock;
	
	time(&now);
	tblock=localtime(&now);
	strftime(timeStampStr,20,"%b %d %H:%M:%S",tblock);
	
	return timeStampStr;
}
/**************************** usage ***************************************/
void usage(void)
{
    printf(
      "\nUsage: sga [X Options] [Options]\n"
      "  Makes animated GIFs\n"
      "\n"
      "  Options:\n"
      "    filename     Open existing file with name filename\n"
      "    -h help      This message\n"
	);
}
/**************************** wmCloseCB ***********************************/
void wmCloseCB(Widget w, XtPointer clientData, XtPointer callData)
{
    UNREFERENCED(clientData);
    UNREFERENCED(callData);

    XtPopdown(w);
}
/**************************** xerrMsg *************************************/
int xerrMsg(const char *fmt, ...)
{
    Widget child;
    va_list vargs;
    static char lstring[FIXED_MSG_SIZE];
    
    va_start(vargs,fmt);
    (void)vsprintf(lstring,fmt,vargs);
    va_end(vargs);
    
    if(lstring[0] != '\0') {
	if(windowMessage) {
	    XBell(display,50); XBell(display,50); XBell(display,50); 
	    cstring=XmStringCreateLtoR(lstring,XmSTRING_DEFAULT_CHARSET);
	    nargs=0;
	    XtSetArg(args[nargs],XmNtitle,"Warning"); nargs++;
	    XtSetArg(args[nargs],XmNmessageString,cstring); nargs++;
#ifndef WIN32
	    warningBox=XmCreateWarningDialog(mainWindow,"warningMessage",
#else
	    warningBox=XmCreateMessageDialog(mainWindow,"warningMessage",
#endif
	      args,nargs);
	    XmStringFree(cstring);
	    child=XmMessageBoxGetChild(warningBox,XmDIALOG_CANCEL_BUTTON);
	    XtDestroyWidget(child);
	    child=XmMessageBoxGetChild(warningBox,XmDIALOG_HELP_BUTTON);
	    XtDestroyWidget(child);
	    XtManageChild(warningBox);
	    XmUpdateDisplay(warningBox);
	}
#ifdef WIN32
	lprintf("%s\n",lstring);
#else
	fprintf(stderr,"%s\n",lstring);
#endif
    }

    return 0;
}
/**************************** xerrorhandler *******************************/
extern "C" int xerrorhandler(Display *dpy, XErrorEvent *event)
{
    char buf[FIXED_MSG_SIZE];

    XGetErrorText(dpy,event->error_code,buf,FIXED_MSG_SIZE);
    sprintf(string,"X Error:\n%s",buf);
#ifdef WIN32
    lprintf("%s\n",string);
#else
    fprintf(stderr,"%s\n",string);
#endif
    if(windowMessage) xerrMsg(string);
    return 0;
}
/**************************** xinfoMsg ************************************/
void xinfoMsg(char *sstring)
{
    Widget child,w;

    if(sstring[0] != '\0') {
	XmString sstringString=XmStringCreateLtoR(sstring,
	  XmSTRING_DEFAULT_CHARSET);
	XmString cancelString=XmStringCreateLocalized("Cancel");
	nargs=0;
	XtSetArg(args[nargs],XmNtitle,"Information"); nargs++;
	XtSetArg(args[nargs],XmNmessageString,sstringString); nargs++;
	XtSetArg(args[nargs],XmNokLabelString,cancelString); nargs++;
#ifndef WIN32
	w=XmCreateInformationDialog(mainWindow,"informationMessage",args,nargs);
#else
	w=XmCreateMessageDialog(mainWindow,"informationMessage",args,nargs);
#endif
	XmStringFree(sstringString);
	XmStringFree(cancelString);
	child=XmMessageBoxGetChild(w,XmDIALOG_CANCEL_BUTTON);
	XtDestroyWidget(child);
	child=XmMessageBoxGetChild(w,XmDIALOG_HELP_BUTTON);
	XtDestroyWidget(child);
	XtManageChild(w);
    }
}
/**************************** xterrorhandler ******************************/
void xterrorhandler(char *message)
{
#ifdef WIN32
    lprintf("%s\n",message);
#else
    fprintf(stderr,"%s\n",message);
#endif
    if(windowMessage) xerrMsg(string);
}
/**************************** executeFSD **********************************/
static void executeFSD(void)
{
    static int response=0;     // Keep it off the stack

  // KE: response is not used here.  It is intended to insure an
  // immediate response.  Since it is not used, the fsdButton could be
  // sent as the client data.  We will leave it for now.
    if(fileSD == NULL) {
	XmString gifDirMask=XmStringCreateLocalized(GIF_DIR_MASK);
	
	nargs=0;
	XtSetArg(args[nargs],XmNdirMask,gifDirMask); nargs++;
	XtSetArg(args[nargs],XmNdialogStyle,
	  XmDIALOG_FULL_APPLICATION_MODAL); nargs++;
	fileSD=XmCreateFileSelectionDialog(mainWindow,"fileSD",args,nargs);
	XtAddCallback(fileSD,XmNokCallback,fileSelectCB,&response);
	XtAddCallback(fileSD,XmNcancelCallback,fileSelectCB,&response);
	XtUnmanageChild(XmFileSelectionBoxGetChild(fileSD,
	  XmDIALOG_HELP_BUTTON));
	XtManageChild(fileSD);
	XmStringFree(gifDirMask);
    } else {
	XtManageChild(fileSD);
    }
}
/**************************** fileSelectCB *********************************/
static void fileSelectCB(Widget w, XtPointer clientData, XtPointer callData)
{
    int *response=(int *)clientData;
    XmFileSelectionBoxCallbackStruct *cbs =
      (XmFileSelectionBoxCallbackStruct *)callData;

    switch(cbs->reason){
    case XmCR_CANCEL:
	XtUnmanageChild(w);
	*response=2;
	break;
    case XmCR_OK:
	if(cbs->value != NULL) {
	    char *fullPathName;
	    int status=SGA_ERROR;
	    
	    *response=1;
	    XmStringGetLtoR(cbs->value,XmFONTLIST_DEFAULT_TAG,&fullPathName);

	  // Do the appropriate action
	    XDefineCursor(display,XtWindow(appShell),watchCursor);
	    XFlush(display);
	    switch(fsdButton) {
	    case FILE_OPEN_BTN:
		status=loadFile(fullPathName);
		break;
	    case FILE_SAVEAS_BTN:
		status=saveFile(fullPathName);
		break;
	    case INSERT_GIF_BTN:
		status=insertFile(fullPathName);
		break;
	    }
	    if(status != SGA_OK) *response=2;
	    XtFree(fullPathName);
	    XtUnmanageChild(w);
	    XUndefineCursor(display,XtWindow(appShell));
	} else {
	    *response=2;
	}
	break;
    default:
	*response=2;
	break;
    }
}
/**************************** insertFile **********************************/
int insertFile(char *fileName)
{
    int error=0;
    int status=SGA_ERROR;

  // If there is no current gif, treat  the insertion as a new file
    if(!gif) {
	status=loadFile(NULL);
	return status;
    }

  // Allocate a new CGif
    CGif *newGif=new CGif();
    if(!newGif) {
	xerrMsg("insertFile: Cannot create space for\n  %s",fileName);
	error=1;
	goto CLEANUP;
    }
    
  // Read the file
    if(fileName) {
	status=newGif->load(fileName);
	if(status != SGA_OK) {
	    xerrMsg("insertFile: Cannot load \n  %s",fileName);
	    error=1;
	    goto CLEANUP;
	}
    }

  // Fix the color table
    XDefineCursor(display,XtWindow(appShell),watchCursor);
    XFlush(display);
    switch(colorTableMergeMode) {
	case MATCH_MODE:
	    status=newGif->matchColors(*gif);
	    if(selectedBlock) selectedBlock->displayData();
	    break;
	case REPLACE_MODE:
	    status=newGif->replaceColors(*gif);
	    refreshColorTableS();
	    if(selectedBlock) selectedBlock->displayData();
	    break;
	case ADD_MODE:
	    status=newGif->addColors(*gif);
	    refreshColorTableS();
	    if(selectedBlock) selectedBlock->displayData();
	    break;
	case ADDREPLACE_MODE:
	    status=newGif->addReplaceColors(*gif);
	    refreshColorTableS();
	    if(selectedBlock) selectedBlock->displayData();
	    break;
    }
    if(status != SGA_OK) {
	xerrMsg("insertFile: Cannot merge the color table for\n  %s\n"
	  "The current Color Table Merge Mode is \"%s\"\n"
	  "Consider changing the Color Table Merge Mode",
	  fileName,
	  colorTableMergeModeString[colorTableMergeMode]);
	error=1;
	goto CLEANUP;
    }

  // Clean up the interface
    gif->markEdited();
    gif->updateListArea();

  // Clean up
  CLEANUP:
    XUndefineCursor(display,XtWindow(appShell));
    if(newGif) {
	delete newGif;
	newGif=NULL;
    }
    
    return error?SGA_OK:SGA_ERROR;
}
/**************************** loadFile ************************************/
int loadFile(char *fileName)
{
    int dirLength;
    int error=0;
    int status;

#if DEBUG_EXC
    print("\nloadFile: Start\n");
#endif

  // If filename is NULL, start a new GIF
  // Otherwise read the file

#if DEBUG_EXC
    print("loadFile: Before Set the window title\n");
    checkHeap(" Check Heap:");
#endif
  // Set the window title
    dirLength=setWindowTitle(fileName);

#if DEBUG_EXC
    print("loadFile: Before Remove any timer\n");
    checkHeap(" Check Heap:");
#endif
  // Remove any timer
    stopPlay();

#if DEBUG_EXC
    print("loadFile: Before Clean up the old file\n");
    checkHeap(" Check Heap:");
#endif
  // Clean up the old file
    if(gif) {
#if DEBUG_EXC
	print("loadFile: Before delete gif\n");
	checkHeap(" Check Heap:");
      //	dumpHeap(" Dump Heap:");
#endif
	delete gif;
#if DEBUG_EXC
	print("loadFile: After delete gif\n");
	checkHeap(" Check Heap:");
      //	dumpHeap(" Dump Heap:");
#endif
	gif=NULL;
#if DEBUG_EXC
	print("loadFile: Before clearDisplayArea\n");
	checkHeap(" Check Heap:");
#endif
	clearDisplayArea("No file loaded");	    
#if DEBUG_EXC
	print("loadFile: Before clearListArea\n");
	checkHeap(" Check Heap:");
#endif
	clearListArea();	    
    }
    selectedBlock=NULL;
    
#if DEBUG_EXC
    print("loadFile: Before Allocate a new CGif\n");
    checkHeap(" Check Heap:");
#endif
  // Allocate a new CGif
    if(fileName) {
	gif=new CGif();
	if(!gif) {
	    xerrMsg("loadFile: Cannot create space for %s",fileName);
	    error=1;
	    goto CLEANUP;
	}
    } else {
	gif=new CGif(defaultNColors);
	if(!gif) {
	    xerrMsg("loadFile: Cannot create space for a new GIF");
	    error=1;
	    goto CLEANUP;
	}
    }
    
#if DEBUG_EXC
    print("loadFile: Before Read the file\n");
    checkHeap(" Check Heap:");
#endif
  // Read the file
    if(fileName) {
	status=gif->load(fileName);
	if(status != SGA_OK) {
	    xerrMsg("loadFile: Cannot load %s\n",fileName);
	    error=1;
	    goto CLEANUP;
	}
    }
    
#if DEBUG_EXC
    print("loadFile: Before Update\n");
    checkHeap(" Check Heap:");
#endif
  // Update
    gif->setGraphicsOptions();	    
    clearDisplayArea("Select a block");	    
    gif->updateListArea();
    XmListSelectPos(listArea,1,True);

  // Clean up
  CLEANUP:
    if(error) {
	if(colorTableS) {
	    currentColorTableIndex=0;
	    XtPopdown(colorTableS);
	}
	if(gif) {
	    delete gif;
	    gif=NULL;
	    clearDisplayArea("No file loaded");	    
	    clearListArea();	    
	}
	XmTextSetString(textArea,"Error loading ");
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),
	  fileName+dirLength);
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),"\n");
	XmTextSetCursorPosition(textArea,0);
	changeDisplayWindow(TEXT_MODE);

	return SGA_ERROR;
    }

  // Update the color table if it is popped up
    refreshColorTableS();
    
#if DEBUG_EXC
    print("loadFile: Before return SGA_OK\n");
    checkHeap(" Check Heap:");
#endif
    return SGA_OK;
}
/**************************** saveFile ************************************/
int saveFile(char *fileName)
{
    int writeOK=0;
    int status;

  // Check if the file exists
    errno=0;
    if(access(fileName,W_OK) == -1) {
	if(errno == ENOENT) {
	  // File does not exist
	    writeOK=1;
	} else {
	  // There is some error
	    char *errstring=strerror(errno);
	    xerrMsg("Error accessing file:\n%s\n%s\n",
	      fileName,errstring);
	    return SGA_ERROR;
	}
    } else {
      // File exists, see whether the user wants to overwrite
	sprintf(string,"Do you want to overwrite file:\n%s",
	  fileName);
	int answer=yesNo(string,NULL,NULL);
	if(answer == 1) writeOK=1;
    }

  // Write the file
    if(gif && writeOK) {
	XDefineCursor(display,XtWindow(appShell),watchCursor);
	XFlush(display);
	status=gif->writeFile(fileName);
      // Update the file name and title bar
	gif->setLongName(fileName);
	if(selectedBlock) selectedBlock->displayData();
	setWindowTitle(fileName);
	XUndefineCursor(display,XtWindow(appShell));
    }

    return SGA_OK;
}
/**************************** listAreaCB **********************************/
static void listAreaCB(Widget w, XtPointer clientData, XtPointer callData)
{
    XmListCallbackStruct *cbs=(XmListCallbackStruct *)callData;
    int pos=cbs->item_position;
    int count=cbs->selected_item_count;
    CBlock *block;

    UNREFERENCED(w);
    UNREFERENCED(clientData);

  // Set the selected block to zero, reset later if everything OK
	selectedBlock=NULL;

  // Don't do anything on unselect
    if(!count) {
	return;
    }

  // Check for a valid gif (probably not necessary)
    if(!gif) {
	XBell(display,50);
	return;
    }

  // Get the associated block
    block=gif->getBlock(pos);
    if(!block) {
	xerrMsg("listAreaCB: Cannot find selected block (%d)",pos);
	return;
    }

  // Run the block's display method
    selectedBlock=block;
    block->displayData();
}
/**************************** yesNoCB *************************************/
int yesNo(char *text, char *ok, char *cancel)
{
    static int response;
    XmString textString=NULL,okString=NULL,cancelString=NULL;
    
  // Create the yes/no dialog if necessary
    if(!yesNoD) {
	nargs=0;
	XtSetArg(args[nargs],XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL);
	  nargs++; 
#ifndef WIN32
	yesNoD=XmCreateQuestionDialog(fileSD,"questionDialog",args,nargs);
#else
	yesNoD=XmCreateMessageDialog(fileSD,"questionDialog",args,nargs);
#endif
	XtUnmanageChild(XmMessageBoxGetChild(yesNoD,XmDIALOG_HELP_BUTTON));
	XtAddCallback(yesNoD,XmNokCallback,yesNoCB,(XtPointer)&response);
	XtAddCallback(yesNoD,XmNcancelCallback,yesNoCB,(XtPointer)&response);
    }
    
  // Set the message and pop it up
    nargs=0;
    if(text) {
	textString=XmStringCreateLtoR(text,XmSTRING_DEFAULT_CHARSET);
	XtSetArg(args[nargs],XmNmessageString,textString); nargs++;
    }
    if(ok) {
	okString=XmStringCreateLtoR(ok,XmSTRING_DEFAULT_CHARSET);
	XtSetArg(args[nargs],XmNokLabelString,okString); nargs++;
    }
    if(cancel) {
	cancelString=XmStringCreateLtoR(cancel,XmSTRING_DEFAULT_CHARSET);
	XtSetArg(args[nargs],XmNcancelLabelString,cancelString); nargs++;
    }
    XtSetValues(yesNoD,args,nargs);
    if(text) XmStringFree(textString);
    if(ok) XmStringFree(okString);
    if(cancel) XmStringFree(cancelString);
    XtManageChild(yesNoD);
    XtPopup(XtParent(yesNoD),XtGrabNone);
    
  // Wait for a response
    response=0;
    while(!response) {
	XtAppProcessEvent(appContext,XtIMAll);
    }

    XtPopdown(XtParent(yesNoD));
    return response;
}
/**************************** yesNoCB *************************************/
static void yesNoCB(Widget w, XtPointer clientData, XtPointer callData)
{
    int *response=(int *)clientData;
    XmAnyCallbackStruct *cbs=(XmAnyCallbackStruct *)callData;

    UNREFERENCED(w);

    switch(cbs->reason) {
    case XmCR_OK:
	*response=1;
	break;
    case XmCR_CANCEL:
	*response=2;
	break;
    }
}
/**************************** changeDisplayWindow *************************/
void changeDisplayWindow(int type)
{
  // Set the current drawing image to NULL in either case
	currentDrawingImage=NULL;
  // Set the scrolled window which is managed
    if(type == TEXT_AREA) {
      // TEXT_AREA
	if(textSW) XtManageChild(textSW);
	if(drawingSW && XtIsManaged(drawingSW)) XtUnmanageChild(drawingSW);
    } else {
      // DRAWING_AREA
	if(drawingSW) XtManageChild(drawingSW);
	if(textSW && XtIsManaged(textSW)) XtUnmanageChild(textSW);
    }
}
/**************************** drawingAreaCB *******************************/
void drawingAreaCB(Widget, XtPointer , XtPointer callData)
{
    XmDrawingAreaCallbackStruct *cbs=(XmDrawingAreaCallbackStruct *)callData;

    if(!currentDrawingImage) return;
    switch(cbs->reason) {
    case XmCR_EXPOSE:
	currentDrawingImage->exposeCB();
	break;
    case XmCR_RESIZE:
	currentDrawingImage->resizeCB();
	break;
    }
}
/**************************** startPlay ***********************************/
void startPlay(void)
{
    if(playTimerID) {
	XtRemoveTimeOut(playTimerID);
	playTimerID=(XtIntervalId)0;
    }
    playTimerID=XtAppAddTimeOut(appContext,0,playTimerCB,
      (XtPointer)NULL);
}
/**************************** stopPlay ************************************/
void stopPlay(void)
{
    if(playTimerID) {
	XtRemoveTimeOut(playTimerID);
	playTimerID=(XtIntervalId)0;
    }
}
/**************************** createDefaultOptionArea *********************/
Widget createDefaultOptionArea(Widget parent)
{
    Widget wtop,w,w1;
    static char logo[]="Simple GIF Animator";
    static Pixmap pixmap;

  // Create the form
    nargs=0;
    XtSetArg(args[nargs],XmNfractionBase,24); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    wtop=XmCreateForm(parent,"defaultOptionArea",args,nargs);

  // Create a row column
    nargs=0;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition, 4); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition, 20); nargs++;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNtopPosition, 9); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNbottomPosition, 15); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    w1=XmCreateRowColumn(wtop,"sgaRC",args,nargs);
    XtManageChild(w1);

  // Get the background pixel
    Pixel bg;
    nargs=0;
    XtSetArg(args[nargs],XmNbackground,&bg); nargs++;
    XtGetValues(w1,args,nargs);

  // Get the logo
    pixmap=getLogo(bg,sgaLogo);
    if(pixmap) {
      // Create a label with pixmap
	nargs=0;
	XtSetArg(args[nargs],XmNlabelType,XmPIXMAP); nargs++;
	XtSetArg(args[nargs],XmNlabelPixmap,pixmap); nargs++;
	XtSetArg(args[nargs],XmNtraversalOn,False); nargs++;
	w=XmCreateLabel(w1,"label",args,nargs);
	XtManageChild(w);
    }
  // Create a label with the logo string
    cstring=XmStringCreateLtoR(logo,XmSTRING_DEFAULT_CHARSET);
    nargs=0;
    XtSetArg(args[nargs],XmNlabelString,cstring); nargs++;
    XtSetArg(args[nargs],XmNtraversalOn,False); nargs++;
    w=XmCreateLabel(w1,"label",args,nargs);
    XmStringFree(cstring);
    XtManageChild(w);

  // Manage this one so the sizes will be set
    XtManageChild(wtop);

    return wtop;
}

/**************************** createGlobalOptionArea **********************/
Widget createGlobalOptionArea(Widget parent)
{
    Widget w,wtop,wrc,wrc1,wrc2;

  // Create the form
    nargs=0;
    wtop=XmCreateForm(parent,"globalOptionArea",args,nargs);

  // Create a row column
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    wrc=XmCreateRowColumn(wtop,"sgaRC",args,nargs);

  // Create the row column for the text entries
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
    wrc1=XmCreateRowColumn(wrc,"sgaRC",args,nargs);

  // Background color index
    createLabelBox(wrc1,"labelBox","Background Color Index",3,"",optCB,
      (XtPointer)BG_COLOR_INDEX_ENTRY,&bgColorIndexEntry);

  // Logical screen width
    createLabelBox(wrc1,"labelBox","Logical Screen Width",5,"",optCB,
      (XtPointer)LOGICAL_SCREEN_WIDTH_ENTRY,&logicalScreenWidthEntry);

  // Logical screen height
    createLabelBox(wrc1,"labelBox","Logical Screen Height",5,"",optCB,
      (XtPointer)LOGICAL_SCREEN_HEIGHT_ENTRY,&logicalScreenHeightEntry);

  // Create the row column for the push buttons entries
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
    wrc2=XmCreateRowColumn(wrc,"sgaRC",args,nargs);

  // Edit color table
    nargs=0;
    w=XmCreatePushButton(wrc2,"Edit Color Table...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_EDIT_BTN);
 
  // Unused colors
    nargs=0;
    w=XmCreatePushButton(wrc2,"Show Unused Colors...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_UNUSED_BTN);

  // Trim color table
    nargs=0;
    w=XmCreatePushButton(wrc2,"Trim Color Table",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_TRIM_BTN);

 // Manage the managers
    XtManageChild(wrc1);
    XtManageChild(wrc2);
    XtManageChild(wrc);

    return wtop;    
}
/**************************** createImageOptionArea ***********************/
Widget createImageOptionArea(Widget parent)
{
    Widget w,wtop,wrc,wrc1,wrc1a,wrc1b,wrc2,wrc2a,wrc2aa;
    int i;

  // Create the form
    nargs=0;
    wtop=XmCreateForm(parent,"imageOptionArea",args,nargs);

  // Create a row column
    nargs=0;
    XtSetArg(args[nargs],XmNmarginHeight,0); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
    wrc=XmCreateRowColumn(wtop,"sgaRC",args,nargs);

  // Create a row column
    nargs=0;
    XtSetArg(args[nargs],XmNmarginHeight,0); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    wrc1=XmCreateRowColumn(wrc,"sgaRC",args,nargs);

  // Create the row column for the text entries
    nargs=0;
    XtSetArg(args[nargs],XmNmarginHeight,0); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
    wrc1a=XmCreateRowColumn(wrc1,"sgaRC",args,nargs);

  // Transparent index
    createLabelBox(wrc1a,"labelBox","Transparent Index",3,"",optCB,
      (XtPointer)TRANSPARENT_INDEX_ENTRY,&transparentIndexEntry);

  // Delay time
    createLabelBox(wrc1a,"labelBox","Delay Time (1/100 sec)",5,"",optCB,
      (XtPointer)DELAY_TIME_ENTRY,&delayTimeEntry);

  // Left offset
    createLabelBox(wrc1a,"labelBox","Left Offset",5,"",optCB,
      (XtPointer)LEFT_OFFSET_ENTRY,&leftOffsetEntry);

  // Top offset
    createLabelBox(wrc1a,"labelBox","Top Offset",5,"",optCB,
      (XtPointer)TOP_OFFSET_ENTRY,&topOffsetEntry);

  // Create the row column for the check boxes
    nargs=0;
    XtSetArg(args[nargs],XmNmarginHeight,0); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
    wrc1b=XmCreateRowColumn(wrc1,"sgaRC",args,nargs);

  // Apply to all
    nargs=0;
    w=XmCreateToggleButton(wrc1b,"Apply To All Images",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNvalueChangedCallback,optCB,
      (XtPointer)APPLY_ALL_BTN);
    applyToAllImagesBtn=w;

  // Separator
    nargs=0;
    XtSetArg(args[nargs],XmNseparatorType,XmNO_LINE); nargs++;
  //    XtSetArg(args[nargs],XmNheight,10); nargs++;
    w=XmCreateSeparator(wrc1b,"sgaSeparator",args,nargs);
    XtManageChild(w);

  // Transparent Color
    nargs=0;
    w=XmCreateToggleButton(wrc1b,"Transparent Color",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNvalueChangedCallback,optCB,
      (XtPointer)TRANSPARENT_COLOR_BTN);
    transparentColorBtn=w;
    
  // Interlace
    nargs=0;
    w=XmCreateToggleButton(wrc1b,"Interlace",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNvalueChangedCallback,optCB,
      (XtPointer)INTERLACE_BTN);
    interlaceBtn=w;
    
  // Local Color table
    nargs=0;
    w=XmCreateToggleButton(wrc1b,"Local Color Table",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNvalueChangedCallback,optCB,
      (XtPointer)LOCAL_COLOR_TABLE_BTN);
    localColorTableBtn=w;

  // Disposal method
  // Create a row column
    nargs=0;
    XtSetArg(args[nargs],XmNmarginHeight,0); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    wrc2=XmCreateRowColumn(wrc,"sgaRC",args,nargs);

  // Create a row column
    nargs=0;
    XtSetArg(args[nargs],XmNmarginHeight,0); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    wrc2a=XmCreateRowColumn(wrc2,"sgaRC",args,nargs);

  // Create a row column
    nargs=0;
    XtSetArg(args[nargs],XmNmarginHeight,0); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    wrc2aa=XmCreateRowColumn(wrc2a,"sgaRC",args,nargs);

  // Create a label
    XmString xmString[4];

    xmString[0]=XmStringCreateLocalized("Disposal Method");
    nargs=0;
    XtSetArg(args[nargs],XmNlabelString,xmString[0]); nargs++;
    XtSetArg(args[nargs],XmNtraversalOn,False); nargs++;
    w=XmCreateLabel(wrc2aa,"label",args,nargs);
    XmStringFree(xmString[0]);
    XtManageChild(w);

  // Create a menu
    for(i=0; i < 4; i++) {
	xmString[i]=XmStringCreateLocalized(CImage::getDisposalMethodString(i));
    }
    nargs=0;
    XtSetArg(args[nargs],XmNbuttons,xmString); nargs++;
    XtSetArg(args[nargs],XmNbuttonCount,4); nargs++;
    XtSetArg(args[nargs],XmNsimpleCallback,disposalMethodCB); nargs++;
    w=XmCreateSimpleOptionMenu(wrc2aa,"optionMenu",args,nargs);
    for(i=0; i < 4; i++) {
	XmStringFree(xmString[i]);
    }
    XtManageChild(w);
    disposalMethodMenu=w;

  // Manage the managers
    XtManageChild(wrc1a);
    XtManageChild(wrc1b);
    XtManageChild(wrc1);
    XtManageChild(wrc2aa);
    XtManageChild(wrc2a);
    XtManageChild(wrc2);
    XtManageChild(wrc);

    return wtop;    
}
/**************************** createCommentOptionArea *********************/
Widget createCommentOptionArea(Widget parent)
{
    Widget wtop,w,w1;

  // Create the form
    nargs=0;
    XtSetArg(args[nargs],XmNfractionBase,3); nargs++;
    w1=wtop=XmCreateForm(parent,"commentOptionArea",args,nargs);

  // Create scrolled text
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNeditMode,XmMULTI_LINE_EDIT); nargs++;
    XtSetArg(args[nargs],XmNrows,6); nargs++;
    XtSetArg(args[nargs],XmNcolumns,50); nargs++;
    w=XmCreateScrolledText(w1,"textArea",args,nargs);
    XmTextSetString(w,"");
    XtManageChild(w);
    commentText=w;

  // Create apply button
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_WIDGET); nargs++;
    XtSetArg(args[nargs],XmNtopWidget,w); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNresizable,False); nargs++;
    w=XmCreatePushButton(w1,"Apply",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,optCB,(XtPointer)COMMENT_APPLY_BTN);
    commentApplyBtn=w;

    return wtop;    
}
/**************************** createColorTablePopup ***********************/
void createColorTablePopup(void)
{
    Widget w;
    Widget pane;
    Widget actionArea,colorArea,editArea;
    Widget controlsRC,redRC,greenRC,blueRC;
    Widget colorF;
    XmString xmString;
    
    if(colorTableS != NULL) {
	return;
    }

  // Initialize the sample color
    sampleColor.red=sampleColor.green=sampleColor.blue=0;
    sampleColor.flags=DoRed|DoGreen|DoBlue;
    if(!XAllocColor(display,DefaultColormap(display,screen),&sampleColor)) {
	xerrMsg("createColorTablePopup: Could not allocate sample color");
	sampleColor.pixel=BlackPixel(display,screen);
    }
    
  // Popup shell
    nargs=0;
    XtSetArg(args[nargs],XmNtitle,"Color Table"); nargs++;
    XtSetArg(args[nargs],XmNdeleteResponse, XmDO_NOTHING); nargs++;
    colorTableS=XtCreatePopupShell("colorTableS",topLevelShellWidgetClass,
      appShell,args,nargs);
    XmAddWMProtocolCallback(colorTableS,WM_DELETE_WINDOW,
      wmCloseCB,NULL);
    
  // Paned window
    nargs=0;
    XtSetArg(args[nargs],XmNsashWidth,1); nargs++;
    XtSetArg(args[nargs],XmNsashHeight,1); nargs++;
    pane=XmCreatePanedWindow(colorTableS,"panel",args,nargs);

  // Color area
    nargs=0;
    XtSetArg(args[nargs],XmNnumColumns,16); nargs++;
    XtSetArg(args[nargs],XmNpacking,XmPACK_COLUMN); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    colorArea=XmCreateRowColumn(pane,"colorArea",args,nargs);

  // Color buttons
    for(int i=0; i < 256; i++) {
	nargs=0;
	XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
	XtSetArg(args[nargs],XmNrecomputeSize,False); nargs++;
	w=XmCreatePushButton(colorArea,"",args,nargs);
	XtManageChild(w);
	XtAddCallback(w,XmNactivateCallback,colorTableCB,(XtPointer)i);
	colorButtons[i]=w;
    }

  // Edit area
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    editArea=XmCreateRowColumn(pane,"editArea",args,nargs);

  // Controls row column
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
    controlsRC=XmCreateRowColumn(editArea,"controlsRC",args,nargs);

  // Index
    createLabelBox(controlsRC,"labelBox","Index",3,"",colorTableCB,
      (XtPointer)COLOR_TABLE_INDEX_ENTRY,&colorTableIndexEntry);

  //Red
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    redRC=XmCreateRowColumn(controlsRC,"sgaRC",args,nargs);

    xmString=XmStringCreateLocalized("Red  ");
    nargs=0;
    XtSetArg(args[nargs],XmNlabelString,xmString); nargs++;
    w=XmCreateLabel(redRC,"sgaLabel",args,nargs);
    XtManageChild(w);
    XmStringFree(xmString);

    nargs=0;
    XtSetArg(args[nargs],XmNminimum,0); nargs++;
    XtSetArg(args[nargs],XmNmaximum,0xff); nargs++;
    XtSetArg(args[nargs],XmNvalue,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    XtSetArg(args[nargs],XmNshowValue,False); nargs++;
    XtSetArg(args[nargs],XmNscaleMultiple,1); nargs++;
    w=XmCreateScale(redRC,"sgaScale",args,nargs);
    XtManageChild(w);
    colorTableRedScale=w;
    XtAddCallback(w,XmNdragCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_RED_SCALE);
    XtAddCallback(w,XmNvalueChangedCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_RED_SCALE);

    nargs=0;
    XtSetArg(args[nargs],XmNcolumns,3); nargs++;
    XtSetArg(args[nargs],XmNmaxLength,3); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,0); nargs++;
    w=XmCreateTextField(redRC,"sgaTextField",args,nargs);
    XtManageChild(w);
    colorTableRedEntry=w;
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_RED_ENTRY);

  //Green
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    greenRC=XmCreateRowColumn(controlsRC,"sgaRC",args,nargs);

    xmString=XmStringCreateLocalized("Green");
    nargs=0;
    XtSetArg(args[nargs],XmNlabelString,xmString); nargs++;
    w=XmCreateLabel(greenRC,"sgaLabel",args,nargs);
    XtManageChild(w);
    XmStringFree(xmString);

    nargs=0;
    XtSetArg(args[nargs],XmNminimum,0); nargs++;
    XtSetArg(args[nargs],XmNmaximum,0xff); nargs++;
    XtSetArg(args[nargs],XmNvalue,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    XtSetArg(args[nargs],XmNshowValue,False); nargs++;
    XtSetArg(args[nargs],XmNscaleMultiple,1); nargs++;
    w=XmCreateScale(greenRC,"sgaScale",args,nargs);
    XtManageChild(w);
    colorTableGreenScale=w;
    XtAddCallback(w,XmNdragCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_GREEN_SCALE);
    XtAddCallback(w,XmNvalueChangedCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_GREEN_SCALE);

    nargs=0;
    XtSetArg(args[nargs],XmNcolumns,3); nargs++;
    XtSetArg(args[nargs],XmNmaxLength,3); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,0); nargs++;
    w=XmCreateTextField(greenRC,"sgaTextField",args,nargs);
    XtManageChild(w);
    colorTableGreenEntry=w;
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_GREEN_ENTRY);

  //Blue
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    blueRC=XmCreateRowColumn(controlsRC,"sgaRC",args,nargs);

    xmString=XmStringCreateLocalized("Blue ");
    nargs=0;
    XtSetArg(args[nargs],XmNlabelString,xmString); nargs++;
    w=XmCreateLabel(blueRC,"sgaLabel",args,nargs);
    XtManageChild(w);
    XmStringFree(xmString);

    nargs=0;
    XtSetArg(args[nargs],XmNminimum,0); nargs++;
    XtSetArg(args[nargs],XmNmaximum,0xff); nargs++;
    XtSetArg(args[nargs],XmNvalue,0); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    XtSetArg(args[nargs],XmNshowValue,False); nargs++;
    XtSetArg(args[nargs],XmNscaleMultiple,1); nargs++;
    w=XmCreateScale(blueRC,"sgaScale",args,nargs);
    XtManageChild(w);
    colorTableBlueScale=w;
    XtAddCallback(w,XmNdragCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_BLUE_SCALE);
    XtAddCallback(w,XmNvalueChangedCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_BLUE_SCALE);

    nargs=0;
    XtSetArg(args[nargs],XmNcolumns,3); nargs++;
    XtSetArg(args[nargs],XmNmaxLength,3); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,0); nargs++;
    w=XmCreateTextField(blueRC,"sgaTextField",args,nargs);
    XtManageChild(w);
    colorTableBlueEntry=w;
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_BLUE_ENTRY);

  // Color form
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
    XtSetArg(args[nargs],XmNfractionBase,100); nargs++;
    colorF=XmCreateForm(editArea,"colorForm",args,nargs);

  // Old color
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNtopPosition,10); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNbottomPosition,45); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition,20); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition,80); nargs++;
    XtSetArg(args[nargs],XmNbackground,WhitePixel(display,screen)); nargs++;
  //    XtSetArg(args[nargs],XmNresizeWidth,False); nargs++;
  //    XtSetArg(args[nargs],XmNresizeHeight,False); nargs++;
    w=XmCreateRowColumn(colorF,"colorSample",args,nargs);
    XtManageChild(w);
    colorTableOldSample=w;

  // New color
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNtopPosition,55); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNbottomPosition,90); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition,20); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition,80); nargs++;
    XtSetArg(args[nargs],XmNbackground,sampleColor.pixel); nargs++;
  //    XtSetArg(args[nargs],XmNresizeWidth,False); nargs++;
  //    XtSetArg(args[nargs],XmNresizeHeight,False); nargs++;
    w=XmCreateRowColumn(colorF,"colorSample",args,nargs);
    XtManageChild(w);
    colorTableNewSample=w;

  // Action area
    nargs=0;
    XtSetArg(args[nargs],XmNfractionBase,5); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNskipAdjust,True); nargs++;
    actionArea=XmCreateForm(pane,"actionArea",args,nargs);

    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition,1); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition,2); nargs++;
    w=XmCreatePushButton(actionArea,"Apply",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_APPLY_BTN);

    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition,3); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition,4); nargs++;
    w=XmCreatePushButton(actionArea,"Close",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,colorTableCB,
      (XtPointer)COLOR_TABLE_CLOSE_BTN);

  // Manage the managers last
    XtManageChild(redRC);
    XtManageChild(greenRC);
    XtManageChild(blueRC);
    XtManageChild(colorF);
    XtManageChild(controlsRC);
    XtManageChild(colorArea);
    XtManageChild(editArea);
    XtManageChild(actionArea);
    XtManageChild(pane);
}
/**************************** popupColorTable *****************************/
void popupColorTable(void)
{
    const Pixel *pixels=NULL;

  // Create the shell if not created
    if(!colorTableS) createColorTablePopup();

  // Get parameters from the GIF
    if(gif) {
	pixels=gif->getPixels();
    }
    
  // Get the background pixel
    Pixel bg;
    nargs=0;
    XtSetArg(args[nargs],XmNbackground,&bg); nargs++;
    XtGetValues(colorTableS,args,nargs);

  // Color buttons
    for(int i=0; i < 256; i++) {
	Widget w=colorButtons[i];
	if(gif && i < gif->getNColors()) {
	    nargs=0;
	    XtSetArg(args[nargs],XmNbackground,pixels[i]); nargs++;
	    XtSetValues(w,args,nargs);
	    XtSetSensitive(w,True);
	} else {
	    nargs=0;
	    XtSetArg(args[nargs],XmNbackground,bg); nargs++;
	    XtSetValues(w,args,nargs);
	    XtSetSensitive(w,False);
	}
    }

  // Initialize the index
    setColorTableIndex(0);
    
  // Pop it up
    XtPopup(colorTableS,XtGrabNone);
}
/**************************** setColorTableIndex **************************/
void setColorTableIndex(int index)
{
    char stringValue[4];
    int red,green,blue;
    Pixel pixel;

    if(gif) {
	int nColors=gif->getNColors();
	if(index < 0) index=0;
	if(index >= nColors) index=nColors;
	red=gif->getRed(index);
	green=gif->getGreen(index);
	blue=gif->getBlue(index);
	const Pixel *pixels=gif->getPixels();
	pixel=pixels[index];
    } else {
	index=red=green=blue=0;
	pixel=BlackPixel(display,screen);
    }

    currentColorTableIndex=index;

  // Index
    sprintf(stringValue,"%d",index);
    XmTextSetString(colorTableIndexEntry,stringValue);

  // Red
    sprintf(stringValue,"%d",red);
    XmTextSetString(colorTableRedEntry,stringValue);
    nargs=0;
    XtSetArg(args[nargs],XmNvalue,red); nargs++;
    XtSetValues(colorTableRedScale,args,nargs);
    sampleColor.red=red<<8;

  // Green
    sprintf(stringValue,"%d",green);
    XmTextSetString(colorTableGreenEntry,stringValue);
    nargs=0;
    XtSetArg(args[nargs],XmNvalue,green); nargs++;
    XtSetValues(colorTableGreenScale,args,nargs);
    sampleColor.green=green<<8;

  // Blue
    sprintf(stringValue,"%d",blue);
    XmTextSetString(colorTableBlueEntry,stringValue);
    nargs=0;
    XtSetArg(args[nargs],XmNvalue,blue); nargs++;
    XtSetValues(colorTableBlueScale,args,nargs);
    sampleColor.blue=blue<<8;

  // Old and new color
    nargs=0;
    XtSetArg(args[nargs],XmNbackground,pixel); nargs++;
    XtSetValues(colorTableOldSample,args,nargs);
    XtSetValues(colorTableNewSample,args,nargs);
}
/**************************** updateColorTableSample **********************/
void updateColorTableSample(int component,int value)
{
    char stringValue[4];
    Pixel blackPixel=BlackPixel(display,screen);
    Colormap colormap=DefaultColormap(display,screen);
    
    
    switch(component) {
    case DO_RED:
	sprintf(stringValue,"%d",value);
	XmTextSetString(colorTableRedEntry,stringValue);
	sampleColor.red=value<<8;
	break;
    case DO_GREEN:
	sprintf(stringValue,"%d",value);
	XmTextSetString(colorTableGreenEntry,stringValue);
	sampleColor.green=value<<8;
	break;
    case DO_BLUE:
	sprintf(stringValue,"%d",value);
	XmTextSetString(colorTableBlueEntry,stringValue);
	sampleColor.blue=value<<8;
	break;
    }
    
  // Set the new color
    if(sampleColor.pixel != blackPixel) {
	XFreeColors(display,colormap,&sampleColor.pixel,1,0);
    }
    if(!XAllocColor(display,colormap,&sampleColor)) {
	sampleColor.pixel=blackPixel;
    }
    nargs=0;
    XtSetArg(args[nargs],XmNbackground,sampleColor.pixel); nargs++;
    XtSetValues(colorTableNewSample,args,nargs);
}
